# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""Public MuyGPyS modules and functions."""

__version__ = "0.7.2"

from MuyGPyS._src.config import (
    config as config,
    jax_config as jax_config,
    MPI as MPI,
)
